import { AxiosRequestConfig, AxiosPromise } from 'axios';
import { M, L, A, I, S, U, O } from 'ts-toolbelt';

interface Options extends AxiosRequestConfig {
    queryMethod?: 'url' | 'body';
}
interface Stringifiable {
    toApicalypseString(): string;
}
interface Builder<T = any> extends Stringifiable {
    queryFields: {
        fields?: string;
        exclude?: string;
        sort?: string;
        limit?: string;
        offset?: string;
        search?: string;
        where: string[];
    };
    queryEndpoint?: string;
    queryName?: string;
}
interface NamedBuilder<T, N extends string> extends Omit<Builder<T>, 'queryEndpoint' | 'queryName'> {
    queryEndpoint: string;
    queryName: N;
}
interface NarrowBuilder<T> extends Builder<T> {
    __narrow: true;
}
interface CountBuilder<T> extends Builder<T> {
    __count: true;
}
type PickWith<T, K extends keyof T, X> = X extends keyof T ? Pick<T, K | X> : Pick<T, K>;
interface Pipe<T extends object, mode extends 'result' | 'count' = 'result'> {
    <A>(...steps: (BuilderOperator<T, T> | BuilderOperatorNarrow<T, A>)[]): Stringifiable & Executor<FallbackIfUnknown<A, PickWith<T, never, 'id'>>, mode>;
}
type PipeSub<T extends object, Ret extends string, mode extends 'result' | 'count' = 'result'> = {
    <A>(...steps: (BuilderOperator<T, T> | BuilderOperatorNarrow<T, A>)[]): NamedBuilder<FallbackIfUnknown<A, PickWith<T, never, 'id'>>, Ret> & (mode extends 'count' ? CountBuilder<any> : {});
};
interface Executor<T, mode extends 'result' | 'count' = 'result'> {
    execute(url: string, options?: Options): AxiosPromise<{
        result: T[];
        count: {
            count: number;
        };
    }[mode]>;
}
type ResultMultiMono<T extends NamedBuilder<any, any>> = {
    name: T["queryName"];
} & (T extends CountBuilder<any> ? {
    count: number;
} : {
    result: T extends NamedBuilder<infer S, any> ? S[] : never;
});
interface ExecutorMulti<T extends Builder<any>[]> {
    execute(url: string, options?: Options): AxiosPromise<T extends (infer S)[] ? ResultMultiMono<S extends NamedBuilder<any, any> ? S : never>[] : never>;
}
interface BuilderOperator<T, R> {
    (builder: Builder<T>): Builder<R>;
}
interface BuilderOperatorNarrow<T, R> {
    (builder: Builder<T>): NarrowBuilder<R>;
}
interface NamedBuilderOperator<T, N extends string> {
    (builder: Builder<T>): NamedBuilder<T, N>;
}
type StandardOperators = '=' | '!=';
type StringOperators = StandardOperators | '~';
type NumbersOperatos = StandardOperators | '>=' | '>' | '<=' | '<';
type AllowedValues = true | false | null;
type GetOp<T> = T extends number ? NumbersOperatos : T extends string ? StringOperators : StandardOperators;
declare enum WhereFlags {
    RAW = 1,
    NUMBER = 1,
    STRING = 2,
    STARTSWITH = 6,
    ENDSWITH = 10,
    CONTAINS = 14
}
declare enum WhereInFlags {
    AND = 32,
    NAND = 48,
    OR = 64,
    NOR = 80,
    EXACT = 128
}
type FallbackIfUnknown<T, F> = unknown extends T ? F : T;
type OuterKeyCast<T, K extends string> = PickOuterKey<K> & UnionKeyOf<T>;
type UnionKeyOf<T> = T extends Array<infer T> ? keyof T : T extends infer T ? keyof T : never;
type DeepPick<T, P extends string> = T extends M.Primitive ? T : T extends Array<infer I> ? DeepPick<I, P>[] : T extends object ? string extends P ? number : PickWith<InnerPick<T, P>, OuterKeyCast<T, P>, 'id'> : never;
type InnerPick<T, K extends string> = {
    [key in keyof T]: DeepPick<T[key], InnerKey<Extract<key, string>, K>>;
};
type FlatPath<O extends any, P extends L.List<A.Key>, X = Path<O, P>> = X extends object ? number : X;
type PickOuterKey<K extends string> = K extends '*' ? string : KeyHead<K>;
type KeyHead<K extends string> = K extends `${infer K}.${string}` ? K : K;
type InnerKey<key extends string, K> = [
    Extract<K, `${key}.${string}`>
] extends [`${key}.${infer K}`] ? K : never;
type FlatKey<O, K extends keyof O> = Exclude<O[K], null | undefined> extends Array<infer I> ? I : O[K];
type Flat<O> = Exclude<O, null | undefined> extends Array<infer I> ? I : O;
type _ExcludePrimitiveKeys<O> = O extends M.Primitive ? Omit<O, keyof O> : O;
type _Path<O, P extends L.List<A.Key>, It extends I.Iteration = I.IterationOf<0>> = {
    0: _Path<Flat<A.At<_ExcludePrimitiveKeys<O>, P[I.Pos<It>]>>, P, I.Next<It>>;
    1: O;
}[A.Extends<I.Pos<It>, L.Length<P>>];
type Path<O extends any, P extends L.List<A.Key>> = _Path<O, P> extends infer X ? A.Cast<X, any> : never;
type Index = number | string;
type KeyToIndex<K extends A.Key, SP extends L.List<Index>> = number extends K ? L.Head<SP> : K & Index;
type MetaPath<O, D extends string, St extends string, SP extends L.List<Index> = [], P extends L.List<Index> = []> = {
    [K in keyof Required<O>]: Exclude<MetaPath<FlatKey<O, K>, D, St, L.Tail<SP>, [...P, KeyToIndex<K, SP>]>, string> | S.Join<[...P, KeyToIndex<K, SP>], D> | ([St] extends [never] ? never : S.Join<[...P, St], D>);
};
type NextPath<OP> = U.Select<O.UnionOf<Exclude<OP, string> & {}>, string>;
type CurrentPath<OP> = U.Select<Exclude<OP, object>, string>;
type ExecPath<A, SP extends L.List<Index>, D extends string, St extends string> = NextPath<Path<MetaPath<A, D, St, SP>, SP>>;
type HintPath<A, P extends string, SP extends L.List<Index>, Exec extends string, D extends string, St extends string> = [Exec] extends [never] ? CurrentPath<Path<MetaPath<A, D, St, SP>, SP>> extends never ? ExecPath<A, L.Pop<SP>, D, St> : CurrentPath<Path<MetaPath<A, D, St, SP>, SP>> : Exec | P;
type _AutoPath<A, P extends string, D extends string, St extends string, SP extends L.List<Index> = S.Split<P, D>> = HintPath<A, P, SP, ExecPath<A, SP, D, St>, D, St>;
type AutoPath<O extends any, P extends string, D extends string = '.', St extends string = '*'> = _AutoPath<O, P, D, St>;
type AllAutoPath<O extends object, P extends L.List<string>> = {
    [K in keyof P]: AutoPath<O, P[K] & string>;
};
type NonEmptyStringList = [string, ...string[]];
type ChosenPaths<P extends NonEmptyStringList> = [string] extends P ? '*' : P[number];

/**
 * Select only given fields.
 * Resulting response type is narrowed to only include specified fields (including nested ones).
 *
 * @example
 * ```ts
 * request().pipe(
 *   fields('*'),                   // select all fields
 *   fields(["name", "genres.id"]), // select specific fields, incuding sub-type
 *   fields(["name", "genres.*"]),  // select specific fields, incuding all fields of sub-type
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#fields}
 * @param f
 */
declare function fields<T extends Record<any, any>, P extends NonEmptyStringList>(f: AllAutoPath<T, P> | '*'): BuilderOperatorNarrow<T, DeepPick<T, ChosenPaths<P>>>;
/**
 * Exclude given fields from selection.
 * Resulting response type is narrowed to exclude specified fields (not compatible with nested fields).
 *
 * @example
 * ```ts
 * request().pipe(
 *   fields('*'),
 *   exclude(["name", "created_at"]), // exclude specific fields
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#exclude}
 * @param exclude
 */
declare function exclude<T extends Record<any, any>, K extends keyof T>(exclude: K[]): BuilderOperatorNarrow<T, Omit<T, K>>;
/**
 * Sort results by specified field.
 *
 * @example
 * ```ts
 * request().pipe(
 *   sort('name'),          // defaults to 'asc' when direction is not specified
 *   sort('name', 'desc'),  // manually specify direction
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#sorting}
 * @param field
 * @param direction
 */
declare function sort<T extends Record<any, any>>(field: keyof T, direction?: 'asc' | 'desc'): BuilderOperator<T, T>;
/**
 * Limit the number of results returned by the query.
 *
 * @example
 * ```ts
 * request().pipe(
 *   limit(42), // Only retrieve 42 elements
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#pagination}
 * @param limit
 */
declare function limit<T extends Record<any, any>>(limit: number): BuilderOperator<T, T>;
/**
 * Start results at a given offset.
 *
 * @example
 * ```ts
 * request().pipe(
 *   offset(50), // Start he results at position 50...
 *   limit(42),  // ...and limit the number of results to 42
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#pagination}
 * @param offset
 */
declare function offset<T extends Record<any, any>>(offset: number): BuilderOperator<T, T>;
/**
 * Search based on name, results are sorted by similarity to the given search string.
 *
 * @example
 * ```ts
 * request().pipe(
 *   search('zelda'),
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#search-1}
 * @param search
 */
declare function search<T extends Record<any, any>>(search: string): BuilderOperator<T, T>;
/**
 * Filters results based on given parameters. All parameters are type validated if `request` is given a type.
 * Also see {@link whereIn}.
 *
 * @example
 * ```ts
 * request<MyType>().pipe(
 *   where('ratings', '>', 4),  // ratings greater than 4
 *   where('ratings', '>=', 4), // ratings greater than or equal 4
 *   where('ratings', '<', 4),  // ratings less than 4
 *   where('ratings', '<=', 4), // ratings less than or equal 4
 *   where('name', '=', 'zelda'),   // name is zelda (case sensitive)
 *   where('name', '~', 'zelda'),   // name is zelda (case insensitive)
 *   where('name', '!=', 'zelda'),  // name is not zelda
 *   where('name', '=', 'zelda', WhereFlags.STARTSWITH),  // name starts with zelda (also works with ~)
 *   where('name', '=', 'zelda', WhereFlags.ENDSWITH),    // name ends with zelda (also works with ~)
 *   where('name', '=', 'zelda', WhereFlags.CONTAINS),    // name contains zelda (also works with ~)
 *   where('release_date', '=', null)   // no release date
 *   where('release_date', '!=', null)  // release date has any value
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#filters}
 * @param key
 * @param op
 * @param value
 * @param flag
 */
declare function where<T extends Record<any, any>, P extends string>(key: AutoPath<T, P, '.', never>, op: GetOp<Path<T, S.Split<P, '.'>>>, value: FlatPath<T, S.Split<P, '.'>> | AllowedValues, flag?: WhereFlags): BuilderOperator<T, T>;
/**
 * Filters results based on given parameters. All parameters are type validated if `request` is given a type.
 * Also see {@link where}.
 *
 * @example
 * ```ts
 * request<MyType>().pipe(
 *   whereIn('genres', [1, 2], WhereInFlags.AND),   // Results whose genres includes 1 and 2
 *   whereIn('genres', [1, 2], WhereInFlags.OR),    // Results whose genres includes 1 or 2
 *   whereIn('genres', [1, 2], WhereInFlags.NAND),  // Results whose genres does not contain both 1 and 2, but can be 1 or 2
 *   whereIn('genres', [1, 2], WhereInFlags.NOR),   // Results whose genres does not contain 1 or does not contain 2
 *   whereIn('genres', [1, 2], WhereInFlags.EXACT), // Results whose exclusive genres are 1 and 2
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#filters}
 * @param key
 * @param values
 * @param flag
 */
declare function whereIn<T extends Record<any, any>, P extends string>(key: AutoPath<T, P, '.', never>, values: FlatPath<T, S.Split<P, '.'>>[], flag?: WhereInFlags | WhereFlags): BuilderOperator<T, T>;
/**
 * Combine multiple {@link where} and {@link whereIn} conditions.
 *
 * @example
 * ```ts
 * // Select results that are on platform 1 and 2, and whose name are either 'zelda' or 'link'
 * request<MyType>().pipe(
 *   and(
 *     whereIn('platforms', [1, 2], WhereInFlags.AND),
 *     or(
 *       where('name', '=', 'zelda'),
 *       where('name', '=', 'link'),
 *     )
 *   )
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#filters}
 * @param operators
 */
declare function and<T extends Record<any, any>>(...operators: BuilderOperator<T, T>[]): BuilderOperator<T, T>;
/**
 * Combine multiple {@link where} and {@link whereIn} conditions.
 *
 * @example
 * ```ts
 * // Select results that are on platform 1 and 2, and whose name are either 'zelda' or 'link'
 * request<MyType>().pipe(
 *   and(
 *     whereIn('platforms', [1, 2], WhereInFlags.AND),
 *     or(
 *       where('name', '=', 'zelda'),
 *       where('name', '=', 'link'),
 *     )
 *   )
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#filters}
 * @param operators
 */
declare function or<T extends Record<any, any>, R>(...operators: BuilderOperator<T, R>[]): BuilderOperator<T, R>;

/**
 * Prepare a request to an apicalypse endpoint.
 *
 * @example
 * ```ts
 * // MyType is optional but recommended, so that results are properly typed
 * request<MyType>().pipe(
 *   fields(["name"]),
 *   sort("created_at", "asc"),
 *   where("created_at", ">",  now),
 * )
 * .execute('http:...')
 * .then(results => ...);
 * ```
 */
declare function request<T extends Record<any, any>, mode extends 'result' | 'count' = 'result'>(): {
    pipe: Pipe<T, mode>;
    sub: <S extends string>(queryEndpoint: string, queryName: S) => {
        pipe: PipeSub<T, S, mode>;
    };
};
/**
 * Prepare a multi-query request to an apicalypse endpoint.
 *
 * @example
 * ```ts
 * multi(
 *   // MyType is optional but recommended, so that results are properly typed
 *   request<MyType>()
 *     .sub('endpoint', 'alias') // mandatory for multi requests
 *     .pipe(
 *       fields(["name"]),
 *       sort("created_at", "asc"),
 *       where("created_at", ">",  now),
 *     ),
 *   request<MyType>()
 *     .sub('endpoint', 'alias2') // mandatory for multi requests
 *     .pipe(
 *       fields(["name"]),
 *       sort("created_at", "asc"),
 *       where("created_at", "<",  now),
 *     )
 * )
 * .execute('http:...')
 * .then(response => ...);
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#multi-query}
 * @param builders
 */
declare function multi<T extends Record<any, any>, B extends Builder<T>[]>(...builders: B): Stringifiable & ExecutorMulti<B>;
/**
 * Narrow multi-query results types.
 *
 * @example
 * ```ts
 * interface MyType {
 *   name: string,
 *   created_at: number
 * }
 *
 * multi(
 *  request<MyType>().sub("games", "latest-games").pipe(
 *    fields(["name"]),
 *    sort("created_at", "desc"),
 *    where("created_at", "<",  now),
 *  ),
 *  request<MyType>().sub("games", "coming-soon").pipe(
 *    fields(["created_at"]),
 *    sort("created_at", "desc"),
 *    where("created_at", ">",  now),
 *  )
 * ).execute('http:...').then(response => {
 *   for (const result of response.data) {
 *     if (isNamed(result, 'latest-games')) {
 *       result.result.name // result is of type `{ name: string }`
 *     } else {
 *       result.result.created_at  // result is of type `{ created_at: number }`
 *     }
 *   }
 * });
 * ```
 * @param builder
 * @param name
 */
declare function isNamed<T extends NamedBuilder<any, string>, S extends string>(builder: ResultMultiMono<T>, name: S): builder is ResultMultiMono<Extract<T, NamedBuilder<any, S>>>;

export { AllAutoPath, AllowedValues, AutoPath, Builder, BuilderOperator, BuilderOperatorNarrow, ChosenPaths, CountBuilder, DeepPick, Executor, ExecutorMulti, FallbackIfUnknown, FlatPath, GetOp, NamedBuilder, NamedBuilderOperator, NarrowBuilder, NonEmptyStringList, NumbersOperatos, Options, Path, Pipe, PipeSub, ResultMultiMono, StandardOperators, StringOperators, Stringifiable, WhereFlags, WhereInFlags, and, exclude, fields, isNamed, limit, multi, offset, or, request, search, sort, where, whereIn };
